==> geofeaturekit-0.1.1.tar.gz: geofeaturekit-0.1.1/geofeaturekit/core/extractor.py <==
"""Core feature extraction module."""

import warnings
import os
import networkx as nx
import geopandas as gpd
from typing import Dict, Any, List, Tuple, Optional
import osmnx as ox
from ..utils.network import download_network
from ..utils.poi import download_pois
from ..utils.network_advanced import download_land_use
from ..utils.area import calculate_area_hectares
from ..utils.formatting import round_float, AREA_DECIMALS
from ..exceptions.errors import GeoFeatureKitError
from .metrics import calculate_all_metrics

class UrbanFeatureExtractor:
    """Extract urban features from locations."""
    
    def __init__(self, radius_meters: int = 500, use_cache: bool = True):
        """Initialize extractor.
        
        Args:
            radius_meters: Analysis radius in meters
            use_cache: Whether to cache downloaded data
            
        Raises:
            GeoFeatureKitError: If radius is invalid
        """
        if radius_meters <= 0:
            raise GeoFeatureKitError("Radius must be positive")
            
        self.radius_meters = radius_meters
        self.use_cache = use_cache
        self.cache_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cache')
        self._cache = {}
        
        if use_cache and not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
    
    def _extract_network_features(self, G: nx.Graph) -> Dict[str, Any]:
        """Extract network features from graph."""
        return {
            'basic_metrics': {
                'total_street_length_meters': sum(d['length'] for _, _, d in G.edges(data=True)),
                'total_intersections': len([n for n, d in G.nodes(data=True) if d['street_count'] > 1]),
                'total_dead_ends': len([n for n, d in G.nodes(data=True) if d['street_count'] == 1]),
                'total_nodes': G.number_of_nodes(),
                'total_street_segments': G.number_of_edges()
            },
            'density_metrics': {},
            'connectivity_metrics': {},
            'street_pattern_metrics': {}
        }
    
    def _extract_poi_features(self, pois: gpd.GeoDataFrame) -> Dict[str, Any]:
        """Extract POI features from GeoDataFrame."""
        return {
            'absolute_counts': {
                'total_points_of_interest': len(pois),
            },
            'density_metrics': {},
            'distribution_metrics': {}
        }
    
    def features_from_location(
        self,
        latitude: float,
        longitude: float,
        radius_meters: Optional[int] = None
    ) -> Dict[str, Any]:
        """Extract urban features for a single location.
        
        Args:
            latitude: Location latitude
            longitude: Location longitude
            radius_meters: Analysis radius in meters. If None, uses default radius
            
        Returns:
            Dictionary containing urban features and metrics
            
        Raises:
            ValueError: If coordinates are invalid
            GeoFeatureKitError: If data download fails or metrics calculation fails
        """
        # Validate coordinates
        if not (-90 <= latitude <= 90):
            raise ValueError(f"Invalid latitude: {latitude}. Must be between -90 and 90.")
        if not (-180 <= longitude <= 180):
            raise ValueError(f"Invalid longitude: {longitude}. Must be between -180 and 180.")
        
        radius = radius_meters or self.radius_meters
        
        # Check cache first
        cache_key = self._get_cache_key(latitude, longitude, radius)
        cached_data = self._load_from_cache(cache_key)
        if cached_data is not None:
            return cached_data
        
        try:
            # Download data
            G = download_network(latitude, longitude, radius)
            pois = download_pois(latitude, longitude, radius)
            land_use = download_land_use(latitude, longitude, radius)
            
            # Calculate area in square meters
            area_sqm = calculate_area_hectares(radius) * 10000
            
            # Use the proper metrics calculation function
            metrics = calculate_all_metrics(G, pois, area_sqm)
            
            # Save to cache
            self._save_to_cache(cache_key, metrics)
            
            return metrics
            
        except Exception as e:
            raise GeoFeatureKitError(f"Failed to extract features: {str(e)}")
    
    def features_from_location_batch(
        self,
        locations: List[Tuple[float, float]],
        radius_meters: Optional[int] = None
    ) -> List[Dict[str, Any]]:
        """Extract urban features for multiple locations.
        
        Args:
            locations: List of (latitude, longitude) tuples
            radius_meters: Analysis radius in meters. If None, uses default radius
            
        Returns:
            List of dictionaries containing urban features and metrics
            
        Raises:
            GeoFeatureKitError: If data download fails or metrics calculation fails
        """
        results = []
        errors = []
        
        for i, (lat, lon) in enumerate(locations):
            try:
                result = self.features_from_location(lat, lon, radius_meters)
                results.append(result)
            except Exception as e:
                errors.append(f"Location {i} ({lat}, {lon}): {str(e)}")
        
        if errors:
            warnings.warn(f"Errors occurred during batch processing:\n" + "\n".join(errors))
        
        return results
    
    def _get_cache_key(self, latitude: float, longitude: float, radius: int) -> str:
        """Generate cache key for a location."""
        return f"{latitude},{longitude},{radius}"
    
    def _load_from_cache(self, key: str) -> Optional[Dict[str, Any]]:
        """Load data from cache if available."""
        if not self.use_cache:
            return None
        return self._cache.get(key)
    
    def _save_to_cache(self, key: str, data: Dict[str, Any]) -> None:
        """Save data to cache."""
        if self.use_cache:
            self._cache[key] = data

==> geofeaturekit-0.2.1.tar.gz: geofeaturekit-0.2.1/geofeaturekit/core/extractor.py <==
"""Core feature extraction module."""

import warnings
import os
import networkx as nx
import geopandas as gpd
from typing import Dict, Any, List, Tuple, Optional
import osmnx as ox
from ..utils.network import download_network
from ..utils.poi import download_pois
from ..utils.network_advanced import download_land_use
from ..utils.area import calculate_area_hectares
from ..utils.formatting import round_float, AREA_DECIMALS
from ..utils.progress import EnhancedProgressTracker, BatchProgressTracker
from ..exceptions.errors import GeoFeatureKitError
from .metrics import calculate_all_metrics

class UrbanFeatureExtractor:
    """Extract urban features from locations."""
    
    def __init__(self, radius_meters: int = 500, use_cache: bool = True, 
                 show_progress: bool = True, progress_detail: str = 'normal'):
        """Initialize extractor.
        
        Args:
            radius_meters: Analysis radius in meters
            use_cache: Whether to cache downloaded data
            show_progress: Whether to show progress bars
            progress_detail: Level of progress detail ('minimal', 'normal', 'verbose')
            
        Raises:
            GeoFeatureKitError: If radius is invalid
        """
        if radius_meters <= 0:
            raise GeoFeatureKitError("Radius must be positive")
            
        self.radius_meters = radius_meters
        self.use_cache = use_cache
        self.show_progress = show_progress
        self.progress_detail = progress_detail
        self.cache_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cache')
        self._cache = {}
        
        if use_cache and not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
    
    def _extract_network_features(self, G: nx.Graph) -> Dict[str, Any]:
        """Extract network features from graph."""
        return {
            'basic_metrics': {
                'total_street_length_meters': sum(d['length'] for _, _, d in G.edges(data=True)),
                'total_intersections': len([n for n, d in G.nodes(data=True) if d['street_count'] > 1]),
                'total_dead_ends': len([n for n, d in G.nodes(data=True) if d['street_count'] == 1]),
                'total_nodes': G.number_of_nodes(),
                'total_street_segments': G.number_of_edges()
            },
            'density_metrics': {},
            'connectivity_metrics': {},
            'street_pattern_metrics': {}
        }
    
    def _extract_poi_features(self, pois: gpd.GeoDataFrame) -> Dict[str, Any]:
        """Extract POI features from GeoDataFrame."""
        return {
            'absolute_counts': {
                'total_points_of_interest': len(pois),
            },
            'density_metrics': {},
            'distribution_metrics': {}
        }
    
    def features_from_location(
        self,
        latitude: float,
        longitude: float,
        radius_meters: Optional[int] = None
    ) -> Dict[str, Any]:
        """Extract urban features for a single location.
        
        Args:
            latitude: Location latitude
            longitude: Location longitude
            radius_meters: Analysis radius in meters. If None, uses default radius
            
        Returns:
            Dictionary containing urban features and metrics
            
        Raises:
            ValueError: If coordinates are invalid
            GeoFeatureKitError: If data download fails or metrics calculation fails
        """
        # Validate coordinates
        if not (-90 <= latitude <= 90):
            raise ValueError(f"Invalid latitude: {latitude}. Must be between -90 and 90.")
        if not (-180 <= longitude <= 180):
            raise ValueError(f"Invalid longitude: {longitude}. Must be between -180 and 180.")
        
        radius = radius_meters or self.radius_meters
        
        # Check cache first
        cache_key = self._get_cache_key(latitude, longitude, radius)
        cached_data = self._load_from_cache(cache_key)
        if cached_data is not None:
            return cached_data
        
        # Initialize progress tracker
        location_desc = f"({latitude:.4f}, {longitude:.4f})"
        progress = EnhancedProgressTracker(
            show_progress=self.show_progress,
            detail_level=self.progress_detail
        )
        
        try:
            progress.start_extraction(location_desc)
            
            # Phase 1: Data Download (70% of total time)
            with progress.phase('download', 'Downloading Data'):
                progress.update_phase_progress(10, "Downloading street network...")
                G = download_network(latitude, longitude, radius)
                
                progress.update_phase_progress(50, "Downloading points of interest...")
                pois = download_pois(latitude, longitude, radius)
                
                progress.update_phase_progress(100, "Downloading land use data...")
                land_use = download_land_use(latitude, longitude, radius)
            
            # Phase 2: Calculate Metrics (25% of total time)
            with progress.phase('calculate', 'Calculating Metrics'):
                progress.update_phase_progress(20, "Calculating area...")
                area_sqm = calculate_area_hectares(radius) * 10000
                
                progress.update_phase_progress(100, "Computing urban metrics...")
                metrics = calculate_all_metrics(G, pois, area_sqm)
            
            # Phase 3: Format Results (5% of total time)
            with progress.phase('format', 'Finalizing Results'):
                progress.update_phase_progress(50, "Formatting data...")
                # Save to cache
                self._save_to_cache(cache_key, metrics)
                progress.update_phase_progress(100, "Complete")
            
            progress.complete_extraction()
            return metrics
            
        except Exception as e:
            progress.error(f"Failed to extract features: {str(e)}")
            raise GeoFeatureKitError(f"Failed to extract features: {str(e)}")
    
    def features_from_location_batch(
        self,
        locations: List[Tuple[float, float]],
        radius_meters: Optional[int] = None
    ) -> List[Dict[str, Any]]:
        """Extract urban features for multiple locations.
        
        Args:
            locations: List of (latitude, longitude) tuples
            radius_meters: Analysis radius in meters. If None, uses default radius
            
        Returns:
            List of dictionaries containing urban features and metrics
            
        Raises:
            GeoFeatureKitError: If data download fails or metrics calculation fails
        """
        results = []
        errors = []
        
        # Initialize batch progress tracker
        batch_progress = BatchProgressTracker(
            total_locations=len(locations),
            show_progress=self.show_progress
        )
        
        batch_progress.start_batch()
        
        for i, (lat, lon) in enumerate(locations):
            location_desc = f"({lat:.4f}, {lon:.4f})"
            batch_progress.start_location(location_desc)
            
            try:
                # Temporarily disable individual progress for batch processing
                original_show_progress = self.show_progress
                self.show_progress = False
                
                result = self.features_from_location(lat, lon, radius_meters)
                results.append(result)
                batch_progress.complete_location(success=True)
                
            except Exception as e:
                errors.append(f"Location {i} ({lat}, {lon}): {str(e)}")
                batch_progress.complete_location(success=False)
            finally:
                # Restore original progress setting
                self.show_progress = original_show_progress
        
        batch_progress.complete_batch()
        
        if errors:
            warnings.warn(f"Errors occurred during batch processing:\n" + "\n".join(errors))
        
        return results
    
    def _get_cache_key(self, latitude: float, longitude: float, radius: int) -> str:
        """Generate cache key for a location."""
        return f"{latitude},{longitude},{radius}"
    
    def _load_from_cache(self, key: str) -> Optional[Dict[str, Any]]:
        """Load data from cache if available."""
        if not self.use_cache:
            return None
        return self._cache.get(key)
    
    def _save_to_cache(self, key: str, data: Dict[str, Any]) -> None:
        """Save data to cache."""
        if self.use_cache:
            self._cache[key] = data
